unit FrustumClass;

interface

Uses OpenGL12;

  type
    FrustumArray = Array[0..5, 0..3] of glFloat;

    TFrustum = Object
    private
      Frustum : FrustumArray;
    public
      procedure CalculateFrustum;
      function PointInFrustum(X, Y, Z : glFloat) : Boolean;
      function SphereInFrustum(X, Y, Z, Radius : glFloat) : Boolean;
      function CubeInFrustum(X, Y, Z, Size : glFloat) : Boolean;
      function BoxInFrustum(X, Y, Z, sizeX, sizeY, sizeZ : glFloat) : Boolean;
    end;

implementation

const
  RIGHT   = 0;            // The RIGHT  side of the frustum
  LEFT    = 1;            // The LEFT   side of the frustum
  BOTTOM  = 2;            // The BOTTOM side of the frustum
  TOP     = 3;            // The TOP    side of the frustum
  BACK    = 4;            // The BACK   side of the frustum
  FRONT   = 5;            // The FRONT  side of the frustum


{------------------------------------------------------------------}
{--- This normalizes a plane (A side) from a given frustum.     ---}
{------------------------------------------------------------------}
procedure NormalizePlane(var frustum : FrustumArray; side : Integer);
var magnitude : glFloat;
begin
  magnitude := sqrt(frustum[side][0] * frustum[side][0] +
                    frustum[side][1] * frustum[side][1] +
                    frustum[side][2] * frustum[side][2]);
  frustum[side][0] := frustum[side][0] / magnitude;
  frustum[side][1] := frustum[side][1] / magnitude;
  frustum[side][2] := frustum[side][2] / magnitude;
  frustum[side][3] := frustum[side][3] / magnitude;
end;


{ TFrustum }
         
{---------------------------------------------------------------------}
{---------------------------------------------------------------------}
procedure TFrustum.CalculateFrustum;
var proj, modl, clip : Array[0..15] of glFloat;
begin
  // glGetFloatv() is used to extract information about our OpenGL world.
  // Below, we pass in GL_PROJECTION_MATRIX to abstract our projection matrix.
  glGetFloatv(GL_PROJECTION_MATRIX, @proj);

  // By passing in GL_MODELVIEW_MATRIX, we can abstract our model view matrix.
  glGetFloatv(GL_MODELVIEW_MATRIX, @modl);

  // Now that we have our modelview and projection matrix, if we combine these 2 matrices,
  // it will give us our clipping planes.  To combine 2 matrices, we multiply them.
  clip[ 0] := modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
  clip[ 1] := modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
  clip[ 2] := modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
  clip[ 3] := modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

  clip[ 4] := modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
  clip[ 5] := modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
  clip[ 6] := modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
  clip[ 7] := modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

  clip[ 8] := modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
  clip[ 9] := modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
  clip[10] := modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
  clip[11] := modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

  clip[12] := modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
  clip[13] := modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
  clip[14] := modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
  clip[15] := modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

  // Now we actually want to get the sides of the frustum.  To do this we take
  // the clipping planes we received above and extract the sides from them.

  // This will extract the RIGHT side of the frustum
  Frustum[RIGHT][0] := clip[ 3] - clip[ 0];
  Frustum[RIGHT][1] := clip[ 7] - clip[ 4];
  Frustum[RIGHT][2] := clip[11] - clip[ 8];
  Frustum[RIGHT][3] := clip[15] - clip[12];

  // Now that we have a normal (A,B,C) and a distance (D) to the plane,
  // we want to normalize that normal and distance.
  NormalizePlane(Frustum, RIGHT);

  // This will extract the LEFT side of the frustum
  Frustum[LEFT][0] := clip[ 3] + clip[ 0];
  Frustum[LEFT][1] := clip[ 7] + clip[ 4];
  Frustum[LEFT][2] := clip[11] + clip[ 8];
  Frustum[LEFT][3] := clip[15] + clip[12];

  // Normalize the LEFT side
  NormalizePlane(Frustum, LEFT);

  // This will extract the BOTTOM side of the frustum
  Frustum[BOTTOM][0] := clip[ 3] + clip[ 1];
  Frustum[BOTTOM][1] := clip[ 7] + clip[ 5];
  Frustum[BOTTOM][2] := clip[11] + clip[ 9];
  Frustum[BOTTOM][3] := clip[15] + clip[13];

  // Normalize the BOTTOM side
  NormalizePlane(Frustum, BOTTOM);

  // This will extract the TOP side of the frustum
  Frustum[TOP][0] := clip[ 3] - clip[ 1];
  Frustum[TOP][1] := clip[ 7] - clip[ 5];
  Frustum[TOP][2] := clip[11] - clip[ 9];
  Frustum[TOP][3] := clip[15] - clip[13];

  // Normalize the TOP side
  NormalizePlane(Frustum, TOP);

  // This will extract the BACK side of the frustum
  Frustum[BACK][0] := clip[ 3] - clip[ 2];
  Frustum[BACK][1] := clip[ 7] - clip[ 6];
  Frustum[BACK][2] := clip[11] - clip[10];
  Frustum[BACK][3] := clip[15] - clip[14];

  // Normalize the BACK side
  NormalizePlane(Frustum, BACK);

  // This will extract the FRONT side of the frustum
  Frustum[FRONT][0] := clip[ 3] + clip[ 2];
  Frustum[FRONT][1] := clip[ 7] + clip[ 6];
  Frustum[FRONT][2] := clip[11] + clip[10];
  Frustum[FRONT][3] := clip[15] + clip[14];

  // Normalize the FRONT side
  NormalizePlane(Frustum, FRONT);
end;


{----------------------------------------------------------------}
{--- This determines if a point is inside of the view frustum ---}
{----------------------------------------------------------------}
function TFrustum.PointInFrustum(X, Y, Z: glFloat): Boolean;
var I : Integer;
begin
  result :=FALSE;
  // Go through all the sides of the frustum
  for I :=0 to 5 do
    // Calculate the plane equation and check if the point is behind a side of the frustum
    if Frustum[i][0] * x + Frustum[i][1] * y + Frustum[i][2] * z + Frustum[i][3] <= 0 then
      exit;  // The point was behind a side, so it ISN'T in the frustum

  // The point was inside of the frustum (In front of ALL the sides of the frustum)
  result :=TRUE;
end;


{----------------------------------------------------------------}
{--- This determines if a sphere is inside our view frustum   ---}
{----------------------------------------------------------------}
function TFrustum.SphereInFrustum(X, Y, Z, Radius: glFloat): Boolean;
var I : Integer;
begin
  result :=FALSE;
  for I :=0 to 5 do
    // Calculate the plane equation and check if the point is behind a side of the frustum
    if Frustum[i][0] * x + Frustum[i][1] * y + Frustum[i][2] * z + Frustum[i][3] <= -radius then
      exit;  // The point was behind a side, so it ISN'T in the frustum

  // The point was inside of the frustum (In front of ALL the sides of the frustum)
  result :=TRUE;
end;


{----------------------------------------------------------------}
{--- This determines if a BOX is in or around our view        ---}
{--- frustum by it's min and max points                       ---}
{----------------------------------------------------------------}
function TFrustum.BoxInFrustum(X, Y, Z, sizeX, sizeY, sizeZ: glFloat): Boolean;
var I : Integer;
begin
  result :=FALSE;
  for I :=0 to 5 do
  begin
    if Frustum[i][0] * x + Frustum[i][1] * y + Frustum[i][2] * z + Frustum[i][3] > 0 then continue;
    if Frustum[i][0] * sizeX + Frustum[i][1] * y + Frustum[i][2] * z + Frustum[i][3] > 0 then continue;
    if Frustum[i][0] * x + Frustum[i][1] * sizeY + Frustum[i][2] * z + Frustum[i][3] > 0 then continue;
    if Frustum[i][0] * sizeX + Frustum[i][1] * sizeY + Frustum[i][2] * z + Frustum[i][3] > 0 then continue;
    if Frustum[i][0] * x + Frustum[i][1] * y + Frustum[i][2] * sizeZ + Frustum[i][3] > 0 then continue;
    if Frustum[i][0] * sizeX + Frustum[i][1] * y + Frustum[i][2] * sizeZ + Frustum[i][3] > 0 then continue;
    if Frustum[i][0] * x + Frustum[i][1] * sizeY + Frustum[i][2] * sizeZ + Frustum[i][3] > 0 then continue;
    if Frustum[i][0] * sizeX + Frustum[i][1] * sizeY + Frustum[i][2] * sizeZ + Frustum[i][3] > 0 then continue;

    // If we get here, it isn't in the frustum
    exit;
  end;

  // Return a true for the box being inside of the frustum
  result :=TRUE;
end;


{----------------------------------------------------------------}
{--- 	This determines if a cube is in or around our view    ---}
{---    frustum by using it's center and 1/2 it's length      ---}
{----------------------------------------------------------------}
function TFrustum.CubeInFrustum(X, Y, Z, Size: glFloat): Boolean;
var I : Integer;
begin
  result :=FALSE;
  for I :=0 to 5 do
  begin
    if Frustum[i][0] * (x - size) + Frustum[i][1] * (y - size) + Frustum[i][2] * (z - size) + Frustum[i][3] > 0 then
       continue;
    if Frustum[i][0] * (x + size) + Frustum[i][1] * (y - size) + Frustum[i][2] * (z - size) + Frustum[i][3] > 0 then
       continue;
    if Frustum[i][0] * (x - size) + Frustum[i][1] * (y + size) + Frustum[i][2] * (z - size) + Frustum[i][3] > 0 then
       continue;
    if Frustum[i][0] * (x + size) + Frustum[i][1] * (y + size) + Frustum[i][2] * (z - size) + Frustum[i][3] > 0 then
       continue;
    if Frustum[i][0] * (x - size) + Frustum[i][1] * (y - size) + Frustum[i][2] * (z + size) + Frustum[i][3] > 0 then
       continue;
    if Frustum[i][0] * (x + size) + Frustum[i][1] * (y - size) + Frustum[i][2] * (z + size) + Frustum[i][3] > 0 then
       continue;
    if Frustum[i][0] * (x - size) + Frustum[i][1] * (y + size) + Frustum[i][2] * (z + size) + Frustum[i][3] > 0 then
       continue;
    if Frustum[i][0] * (x + size) + Frustum[i][1] * (y + size) + Frustum[i][2] * (z + size) + Frustum[i][3] > 0 then
       continue;

    // If we get here, it isn't in the frustum
    exit;
  end;

  result :=TRUE;
end;


end.
